import asyncio
import random
import uuid

from ...AudioSource import AudioData
from ...fetcher import LyricsFetcher
from ...stat import getStat


def need_data(*keys):
    def decorator(func):
        async def wrapper(self, Data, **kwargs):
            if not Data:
                payload = {
                    "op": func.__name__,
                    "d": {"BAD_REQUEST": "This event needs `d`."},
                }
                return await self.sendJson(payload)

            if keys:
                NeedKeys = [key for key in keys if not key in Data.keys()]
                if NeedKeys:
                    payload = {
                        "op": func.__name__,
                        "d": {'"BAD_REQUEST': f"This event needs `{NeedKeys[0]}`."},
                    }
                    return await self.sendJson(payload)

            return await func(self, Data, **kwargs)

        return wrapper

    return decorator


def need_manager(func):
    def wrapper(self, *args, **kwargs):
        if not self.AudioManager:
            payload = {"op": func.__name__, "d": {"NOT_IDENTIFIED": "Identify first."}}

            return self.sendJson(payload)
        return func(self, *args, **kwargs)

    return wrapper


class WebsocketEvents:
    async def HEARTBEAT(self, Data):
        payload = {"op": "HEARTBEAT_ACK", "d": Data}

        await self.sendJson(payload)

    async def GET_STAT(self, Data):
        payload = {"op": "STAT", "d": getStat()}

        payload["d"]["TotalPlayers"] = (
            len(self.AudioManager.voiceClients) if self.AudioManager else 0
        )

        await self.sendJson(payload)

    @need_data("user_id")
    async def IDENTIFY(self, Data):
        if self.AudioManager:
            payload = {
                "op": "IDENTIFY",
                "d": {"ALREADY_IDENTIFIED": "This connection already identified."},
            }
        else:
            await self.initialize_manager(Data["user_id"])
            payload = {"op": "IDENTIFIED", "d": "AudioManager initialized."}

        await self.sendJson(payload)

    @need_manager
    @need_data()
    async def DISCORD_EVENT(self, Data):
        self.AudioManager.discordDispatch(Data)

    @need_manager
    @need_data("guild_id", "volume")
    async def setVolume(self, Data):
        if not Data["volume"] or not (
            isinstance(Data["volume"], int) or Data["volume"].isdigit()
        ):
            payload = {
                "op": "setVolume",
                "d": {"BAD_REQUEST": "`Volume` must be intenger.."},
            }
            return await self.sendJson(payload)

        value = int(Data["volume"])

        if value < 0 or value > 100:
            payload = {
                "op": "setVolume",
                "d": {"BAD_REQUEST": "`Volume` must be `0~200`.."},
            }
            return await self.sendJson(payload)

        self.AudioManager.setVolume(Data["guild_id"], value / 100)

        payload = {
            "op": "setVolume",
            "d": {"guild_id": Data["guild_id"], "volume": Data["volume"]},
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id", "crossfade")
    async def setCrossfade(self, Data):
        try:
            value = float(Data["crossfade"])
        except ValueError:
            payload = {
                "op": "setCrossfade",
                "d": {"BAD_REQUEST": "`crossfade` must be int or float.."},
            }
            return await self.sendJson(payload)

        self.AudioManager.setCrossfade(Data["guild_id"], value)

        payload = {
            "op": "setCrossfade",
            "d": {"guild_id": Data["guild_id"], "crossfade": Data["crossfade"]},
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id", "autoplay")
    async def setAutoplay(self, Data):
        if not isinstance(Data["autoplay"], bool):
            payload = {
                "op": "setAutoplay",
                "d": {"BAD_REQUEST": "`autoplay` must be bool.."},
            }
            return await self.sendJson(payload)

        self.AudioManager.setAutoplay(Data["guild_id"], Data["autoplay"])

        payload = {
            "op": "setAutoplay",
            "d": {"guild_id": Data["guild_id"], "autoplay": Data["autoplay"]},
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id", "filter")
    async def setFilter(self, Data):
        if not isinstance(Data["filter"], dict):
            payload = {
                "op": "setFilter",
                "d": {"BAD_REQUEST": "`filter` must be json.."},
            }
            return await self.sendJson(payload)

        self.AudioManager.getVC(Data["guild_id"]).filter = Data["filter"]

        payload = {
            "op": "setFilter",
            "d": {"guild_id": Data["guild_id"], "filter": Data["filter"]},
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id", "query")
    async def loadSong(self, Data):
        await self.AudioManager.loadSong(Data["guild_id"], Data["query"])

    @need_manager
    @need_data("guild_id", "song")
    async def putSong(self, Data):
        if not isinstance(Data["offset"], dict):
            payload = {
                "op": "putSong",
                "d": {"BAD_REQUEST": "`song` must be dict.."},
            }
            return await self.sendJson(payload)

        Song = AudioData.fromDict(Data["song"])

        await self.AudioManager.putSong(Data["guild_id"], Song)

    @need_manager
    @need_data("guild_id")
    async def skip(self, Data):
        offset = (
            int(Data["offset"])
            if "offset" in Data
            and (isinstance(Data["offset"], int) or Data["offset"].isdigit())
            else 1
        )
        self.AudioManager.skip(Data["guild_id"], offset)

        payload = {
            "op": "skip",
            "d": {
                "guild_id": Data["guild_id"],
                "remain": len(self.AudioManager.getVC(Data["guild_id"]).Queue),
            },
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id")
    async def seek(self, Data):
        if not isinstance(Data["offset"], (int, float)):
            payload = {
                "op": "seek",
                "d": {"BAD_REQUEST": "`offset` must be int or float.."},
            }
            return await self.sendJson(payload)

        self.AudioManager.seek(Data["guild_id"], Data["offset"])

        payload = {
            "op": "seek",
            "d": {"guild_id": Data["guild_id"], "offset": Data["offset"]},
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id")
    async def getQueue(self, Data):
        vc = self.AudioManager.getVC(Data["guild_id"])

        payload = {
            "op": "Queue",
            "d": {
                "guild_id": Data["guild_id"],
                "entries": [Item.toDict() for Item in vc.Queue],
            },
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id")
    async def getState(self, Data):
        vc = self.AudioManager.getVC(Data["guild_id"])

        payload = {
            "op": "State",
            "d": {
                "guild_id": Data["guild_id"],
                "state": vc.state,
                "current": vc.player.current.toDict() if vc.player.current else None,
                "position": {
                    "duration": vc.player.current.duration,
                    "remain": vc.player.current.remain,
                },
                "options": {
                    "autoplay": vc.autoplay,
                    "volume": int(vc.volume * 100),
                    "crossfade": vc.crossfade,
                    "filter": vc.filter,
                },
            },
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id")
    async def pause(self, Data):
        vc = self.AudioManager.getVC(Data["guild_id"])

        vc.pause()

        payload = {
            "op": "pause",
            "d": {"guild_id": Data["guild_id"], "state": vc.state},
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id")
    async def resume(self, Data):
        vc = self.AudioManager.getVC(Data["guild_id"])

        vc.resume()

        payload = {
            "op": "resume",
            "d": {"guild_id": Data["guild_id"], "state": vc.state},
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id")
    async def changePause(self, Data):
        vc = self.AudioManager.getVC(Data["guild_id"])

        vc.changePause()

        payload = {
            "op": "changePause",
            "d": {"guild_id": Data["guild_id"], "state": vc.state},
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id", "repeat")
    async def repeat(self, Data):
        if not isinstance(Data["repeat"], bool):
            payload = {"op": "repeat", "d": {"BAD_REQUEST": "`repeat` must be bool.."}}
            return await self.sendJson(payload)

        vc = self.AudioManager.getVC(Data["guild_id"])

        vc.repeat = Data["repeat"]

        payload = {
            "op": "repeat",
            "d": {"guild_id": Data["guild_id"], "repeat": vc.repeat},
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id")
    async def shuffle(self, Data):
        vc = self.AudioManager.getVC(Data["guild_id"])

        vc.shuffle()

        payload = {
            "op": "shuffle",
            "d": {
                "guild_id": Data["guild_id"],
                "entries": [Item.toDict() for Item in vc.Queue],
            },
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id", "index")
    async def remove(self, Data):
        if not isinstance(Data["index"], (int)):
            payload = {"op": "remove", "d": {"BAD_REQUEST": "`index` must be int.."}}
            return await self.sendJson(payload)

        vc = self.AudioManager.getVC(Data["guild_id"])

        removed = vc.Queue[Data["index"]]
        del vc.Queue[Data["index"]]

        payload = {
            "op": "remove",
            "d": {
                "guild_id": Data["guild_id"],
                "removed": removed.toDict(),
                "entries": [Item.toDict() for Item in vc.Queue],
            },
        }
        return await self.sendJson(payload)

    @need_manager
    @need_data("guild_id")
    async def VC_DESTROY(self, Data):
        self.AudioManager.delVC(Data["guild_id"])

    @need_manager
    @need_data("guild_id", "language")
    async def requestLyrics(self, Data):
        vc = self.AudioManager.getVC(Data["guild_id"])

        if not vc.player.current:
            payload = {
                "op": "requestLyrics",
                "d": {
                    "guild_id": Data["guild_id"],
                    "NotPlaying": "There is no current.",
                },
            }
            return await self.sendJson(payload)

        if not Data["language"] in vc.player.current.lyrics:
            payload = {
                "op": "requestLyrics",
                "d": {
                    "guild_id": Data["guild_id"],
                    "NoLyrics": f"There is no lyrics in {Data['language']}",
                },
            }
            return await self.sendJson(payload)

        current = vc.player.current
        _identify_token = str(uuid.uuid4())
        lyricsLoader = await LyricsFetcher.srv1.load(
            vc.player.current.lyrics[Data["language"]]
        )

        payload = {
            "op": "requestLyrics",
            "d": {
                "guild_id": Data["guild_id"],
                "identify": _identify_token,
                "language": Data["language"],
            },
        }
        await self.sendJson(payload)

        Previous = Now = Next = ""
        Elements = list(lyricsLoader.TextElements.values())
        while not lyricsLoader.is_done and vc.player.current == current:
            Element = lyricsLoader.seek(current.duration)

            if Element and Element["markdown"] and Now != Element["markdown"]:
                NextElements = Elements[Elements.index(Element) + 1 :]

                Previous = Now
                Now = Element["markdown"]

                Next = None
                for NextElement in NextElements:
                    if NextElement["markdown"] != Now:
                        Next = NextElement["markdown"]
                        break

                payload = {
                    "op": "Lyrics",
                    "d": {
                        "guild_id": Data["guild_id"],
                        "identify": _identify_token,
                        "previous": Previous,
                        "current": Now,
                        "next": Next,
                    },
                }
                await self.sendJson(payload)

            await asyncio.sleep(0.1)

        payload = {
            "op": "lyricsDone",
            "d": {
                "guild_id": Data["guild_id"],
                "identify": _identify_token,
                "language": Data["language"],
            },
        }
        return await self.sendJson(payload)



import argparse
import asyncio
import logging
import os
import sys

import colorlog

from .node import server

log = logging.getLogger("discodo")


class loggingFilter(logging.Filter):
    def __init__(self, level):
        self.level = level

    def filter(self, record):
        return record.levelno < self.level


stdoutHandler = logging.StreamHandler(sys.stdout)
stderrHandler = logging.StreamHandler(sys.stderr)
stdoutHandler.addFilter(loggingFilter(logging.WARNING))


def setLoggingLevel(level):
    for logger in [log, logging.getLogger("libav")]:
        addLoggingHandler(logger)
        logger.setLevel(logging.DEBUG)
        stdoutHandler.setLevel(level)
        stderrHandler.setLevel(max(level, logging.WARNING))


ColoredFormatter = colorlog.ColoredFormatter(
    "[%(asctime)s] [%(filename)s] [%(name)s:%(bold)s%(module)s%(reset)s] %(log_color)s[%(levelname)s]%(reset)s: %(message)s",
    datefmt=None,
    reset=True,
    log_colors={
        "DEBUG": "green",
        "INFO": "cyan",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "red,bg_white",
    },
    secondary_log_colors={},
    style="%",
)
stdoutHandler.setFormatter(ColoredFormatter)
stderrHandler.setFormatter(ColoredFormatter)


def addLoggingHandler(logger):
    logger.addHandler(stdoutHandler)
    logger.addHandler(stderrHandler)


parser = argparse.ArgumentParser("discodo")

webGroup = parser.add_argument_group("Webserver Option")

webGroup.add_argument(
    "--host",
    "-H",
    type=str,
    default="0.0.0.0",
    help="the hostname to listen on (default: 0.0.0.0)",
)
webGroup.add_argument(
    "--port",
    "-P",
    type=int,
    default=8000,
    help="the port of the webserver (default: 8000)",
)
webGroup.add_argument(
    "--auth",
    "-A",
    type=str,
    default="hellodiscodo",
    help="the password of the webserver (default: hellodiscodo)",
)
webGroup.add_argument(
    "--ws-interval",
    type=int,
    default=15,
    help="heartbeat interval between discodo server and client (default: 15)",
)
webGroup.add_argument(
    "--ws-timeout",
    type=int,
    default=60,
    help="seconds to close connection there is no respond from client (default: 60)",
)

playerGroup = parser.add_argument_group("Player Option")

playerGroup.add_argument(
    "--default-volume",
    type=int,
    default=100,
    help="player's default volume (default: 100)",
)
playerGroup.add_argument(
    "--default-crossfade",
    type=float,
    default=10.0,
    help="player's default crossfade seconds (default: 10.0)",
)
playerGroup.add_argument(
    "--default-autoplay",
    type=bool,
    default=True,
    help="player's default auto related play state (default: True)",
)
playerGroup.add_argument(
    "--bufferlimit",
    type=int,
    default=5,
    help="seconds of audio will be loaded in buffer (default: 5)",
)
playerGroup.add_argument(
    "--preload",
    type=int,
    default=10,
    help="seconds to load next song before this song ends (default: 10)",
)
playerGroup.add_argument(
    "--timeout",
    type=int,
    default=300,
    help="seconds to cleanup player when connection of discord terminated (default: 300)",
)

logParser = parser.add_argument_group("Logging Option")

logParser.add_argument(
    "--verbose", "-v", action="store_true", help="Print various debugging information"
)

args = parser.parse_args()

if args.verbose:
    setLoggingLevel(logging.DEBUG)
else:
    setLoggingLevel(logging.INFO)

os.environ["VCTIMEOUT"] = str(args.timeout)
os.environ["DEFAULTVOLUME"] = str(round(args.default_volume / 100, 3))
os.environ["DEFAULTCROSSFADE"] = str(args.default_crossfade)
os.environ["DEFAULAUTOPLAY"] = "1" if args.default_autoplay else "0"
os.environ["PRELOAD_TIME"] = str(args.preload)
os.environ["WSINTERVAL"] = str(args.ws_interval)
os.environ["WSTIMEOUT"] = str(args.ws_timeout)
os.environ["AUDIOBUFFERLIMIT"] = str(args.bufferlimit)
os.environ["PASSWORD"] = str(args.auth)


loop = asyncio.get_event_loop()
loop.create_task(
    server.create_server(host=args.host, port=args.port, return_asyncio_server=True)
)
loop.run_forever()



import os
import random
from logging import getLogger

from youtube_related import preventDuplication as relatedClient

from .AudioSource import AudioData, AudioSource
from .player import Player
from .utils import EventEmitter
from .voice_connector import VoiceConnector

log = getLogger("discodo.VoiceClient")

DEFAULTVOLUME = float(os.getenv("DEFAULTVOLUME", "1.0"))
DEFAULTCROSSFADE = float(os.getenv("DEFAULTCROSSFADE", "10.0"))
DEFAULAUTOPLAY = True if os.getenv("DEFAULAUTOPLAY", "1") == "1" else False


class VoiceClient(VoiceConnector):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.relatedClient = relatedClient()

        self.event = EventEmitter()
        self.event.onAny(self.onAnyEvent)
        self.event.on("NeedNextSong", self.getNext)
        self.event.on("SongEnd", self.getNext)

        self.InternalQueue = []
        self._filter = {}

        self.player = None
        self.paused = False
        self._repeat = False

        self.autoplay = DEFAULAUTOPLAY
        self._volume = DEFAULTVOLUME
        self._crossfade = DEFAULTCROSSFADE

        self.event.dispatch("VC_CREATED")

    @property
    def Queue(self):
        """Read only, if you want to edit queue use InternalQueue."""
        return (
            self.InternalQueue[1:]
            if self.InternalQueue and len(self.InternalQueue) > 1
            else []
        )

    def onAnyEvent(self, event, *args, **kwargs):
        self.client.event.dispatch(self.guild_id, event, *args, **kwargs)

    async def getNext(self, **kwargs):
        current = list(kwargs.values()).pop()

        if self.repeat:
            self.InternalQueue.append(await self.getSong(current["webpage_url"]))

        if self.autoplay and (
            not self.InternalQueue
            or (
                self.InternalQueue[0].toDict() == current
                and len(self.InternalQueue) <= 1
            )
        ):
            Related = await self.relatedClient.async_get(current["webpage_url"])
            await self.loadSong(Related["id"])

    def __del__(self):
        guild_id = int(self.guild_id) if self.guild_id else None

        log.info(f"destroying voice client of {guild_id}.")

        self.event.dispatch("VC_DESTROYED")

        if self.client.voiceClients.get(guild_id) == self:
            del self.client.voiceClients[guild_id]

        super().__del__()

        if self.player and self.player.is_alive():
            self.player.stop()

        for Item in self.InternalQueue:
            if isinstance(Item, AudioSource):
                self.loop.call_soon_threadsafe(Item.cleanup)

    async def createSocket(self, data: dict = None):
        await super().createSocket(data)

        if not self.player:
            self.player = Player(self)
            self.player.start()
        else:
            await self.ws.speak(True)

    def putSong(self, Data: AudioData) -> int:
        if not isinstance(Data, (list, AudioData, AudioSource)):
            raise ValueError

        if not isinstance(Data, list):
            Data = [Data]

        for Item in Data:
            self.InternalQueue.append(Item)

        self.event.dispatch(
            "putSong",
            songs=[
                dict(Item.toDict(), index=self.InternalQueue.index(Item))
                for Item in Data
            ],
        )

        return (
            self.InternalQueue.index(Data[0])
            if len(Data) == 1
            else [self.InternalQueue.index(Item) for Item in Data]
        )

    async def loadSong(self, Query: str) -> AudioData:
        Data = await AudioData.create(Query) if isinstance(Query, str) else Query

        self.event.dispatch(
            "loadSong",
            song=(
                [Item.toDict() for Item in Data]
                if isinstance(Data, list)
                else Data.toDict()
            ),
        )

        self.putSong(Data)

        return Data

    def seek(self, offset: int):
        if not self.player.current:
            raise ValueError

        self.player.current.seek(offset)

    def skip(self, offset: int = 1):
        if not self.player.current:
            raise ValueError

        if len(self.InternalQueue) < offset:
            raise ValueError

        del self.InternalQueue[1 : (offset - 1)]

        self.player.current.stop()

    def pause(self) -> bool:
        self.paused = True
        return self.paused

    def resume(self) -> bool:
        self.paused = False
        return self.paused

    def changePause(self) -> bool:
        if self.paused:
            return self.resume()
        else:
            return self.pause()

    def shuffle(self):
        if not self.InternalQueue:
            raise ValueError

        self.InternalQueue = self.InternalQueue[:1] + random.sample(
            self.Queue, k=len(self.Queue)
        )
        return self.Queue

    @property
    def volume(self) -> float:
        return self._volume

    @volume.setter
    def volume(self, value: float):
        self._volume = round(max(value, 0.0), 2)

    @property
    def crossfade(self) -> float:
        return self._crossfade

    @crossfade.setter
    def crossfade(self, value: float):
        self._crossfade = round(max(value, 0.0), 1)

    @property
    def filter(self) -> dict:
        return self._filter

    @filter.setter
    def filter(self, value: dict):
        self._filter = value

    @property
    def state(self) -> str:
        if not self.InternalQueue:
            return "stopped"
        if self.paused:
            return "paused"
        return "playing"

    @property
    def repeat(self) -> bool:
        return self._repeat

    @repeat.setter
    def repeat(self, value: bool):
        self._repeat = value



import asyncio
from itertools import chain
from logging import getLogger

import discord

from .exceptions import VoiceClientNotFound
from .node.client import Node as OriginNode

log = getLogger("discodo.DPYClient")


class NodeClient(OriginNode):
    def __init__(self, DPYClient, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.DPYClient = DPYClient

    async def destroy(self, *args, **kwargs):
        log.infmo(f"destroying Node {self.URL}")
        await super().destroy(*args, **kwargs)

        if self in self.DPYClient.Nodes:
            self.DPYClient.Nodes.remove(self)


class DPYClient:
    def __init__(self, client):
        self.client = client
        self.loop = client.loop or asyncio.get_event_loop()

        self.Nodes = []
        self.__register_event()

    def __register_event(self):
        if hasattr(self.client, "on_socket_response"):
            originFunc = self.client.on_socket_response
        else:
            originFunc = None

        @self.client.event
        async def on_socket_response(*args, **kwargs):
            self.loop.create_task(self.discord_socket_response(*args, **kwargs))

            if originFunc:
                return await originFunc()

    async def discord_socket_response(self, payload):
        if payload["t"] == "VOICE_SERVER_UPDATE":
            VC = self.getVC(payload["d"]["guild_id"])
            SelectNodes = [VC.Node] if VC else [self.getBestNode()]
        else:
            SelectNodes = self.Nodes

        if SelectNodes:
            await asyncio.wait(
                [Node.discordDispatch(payload) for Node in SelectNodes],
                return_when="ALL_COMPLETED",
            )

    def register_node(self, *args, **kwargs):
        return self.loop.create_task(self._register_event(*args, **kwargs))

    async def _register_event(self, *args, **kwargs):
        await self.client.wait_until_ready()

        kwargs["user_id"] = self.client.user.id
        Node = NodeClient(self, *args, **kwargs)
        log.info(f"registering Node {Node.URL}")

        self.Nodes.append(Node)

        Node.emitter.on("VC_DESTROYED", self._vc_destroyed)

        return self

    async def _vc_destroyed(self, Data):
        guild = self.client.get_guild(int(Data["guild_id"]))
        ws = self.__get_websocket(guild.shard_id)

        await ws.voice_state(guild.id, None)

    def getBestNode(self):
        SortedWithPerformance = sorted(
            [Node for Node in self.Nodes if Node.connected.is_set()],
            key=lambda Node: len(Node.voiceClients),
        )

        return SortedWithPerformance[0]

    @property
    def voiceClients(self):
        return {
            ID: Value
            for ID, Value in list(
                chain.from_iterable(
                    [
                        Node.voiceClients.items()
                        for Node in self.Nodes
                        if Node.connected.is_set()
                    ]
                )
            )
        }

    def getVC(self, guild):
        if isinstance(guild, discord.Guild):
            guild = guild.id

        return self.voiceClients.get(int(guild))

    def __get_websocket(self, id):
        if isinstance(self.client, discord.AutoShardedClient):
            return self.client.shards[id].ws
        elif not self.client.shard_id or self.client.shard_id == id:
            return self.client.ws

    async def connect(self, channel):
        log.info(f"connecting to {channel.id} of {channel.guild.id}")
        if not hasattr(channel, "guild"):
            raise ValueError

        ws = self.__get_websocket(channel.guild.shard_id)

        await ws.voice_state(channel.guild.id, channel.id)

    async def disconnect(self, guild):
        log.info(f"disconnecting voice of {guild.id} without destroying")
        ws = self.__get_websocket(guild.shard_id)

        await ws.voice_state(guild.id, None)

    async def destroy(self, guild):
        log.info(f"destroying voice client of {guild.id}")
        if not guild.id in self.voiceClients:
            raise VoiceClientNotFound

        vc = self.getVC(guild.id)
        ws = self.__get_websocket(guild.shard_id)

        await ws.voice_state(guild.id, None)
        await vc.destroy()



import os
import threading

import av

from ..utils.threadLock import withLock
from .AudioFilter import AudioFilter

AVOption = {
    "err_detect": "ignore_err",
    "reconnect": "1",
    "reconnect_streamed": "1",
    "reconnect_delay_max": "5",
}

SAMPLING_RATE = int(os.getenv("SAMPLING_RATE", "48000"))
CHANNELS = int(os.getenv("CHANNELS", "2"))


class Loader:
    def __init__(self, Source: str, AudioFifo: av.AudioFifo):
        self.Source = Source

        self._end = threading.Event()
        self._haveToReloadResampler = threading.Event()
        self._buffering = threading.Lock()
        self._seeking = threading.Lock()

        self.StreamConainer = None
        self.Filter = {}
        self._Filter = {}
        self.FilterGraph = None
        self.reloadResampler()
        self.AudioFifo = AudioFifo

        self.current = 0.0

    def start(self):
        self.BufferLoader = BufferLoader(self)
        self.BufferLoader.start()

    @property
    def duration(self) -> float:
        return (
            round(self.StreamConainer.duration / 1000000)
            if self.StreamConainer
            else None
        )

    def seek(self, offset: float, *args, **kwargs):
        with withLock(self._seeking):
            if not self.StreamConainer and not self._buffering.locked():
                self.StreamConainer = av.open(self.Source, options=AVOption)
            elif not self.StreamConainer:
                while not self.StreamConainer:
                    pass

            self.StreamConainer.seek(round(offset * 1000000), *args, **kwargs)
            self.reload()

    def reload(self):
        self.reloadResampler()

        if not self._buffering.locked():
            if self._end.is_set():
                self._end.clear()
            self.start()

    def reloadResampler(self):
        self._haveToReloadResampler.set()

    def stop(self):
        self._end.set()


class BufferLoader(threading.Thread):
    def __init__(self, Loader):
        threading.Thread.__init__(self)
        self.daemon = True

        self.Loader = Loader
        self.Resampler = None
        self._PreviousFilter = None

    def _do_run(self):
        with withLock(self.Loader._buffering):
            if not self.Loader.StreamConainer:
                self.Loader.StreamConainer = av.open(
                    self.Loader.Source, options=AVOption
                )

            self.Loader.selectAudioStream = self.Loader.StreamConainer.streams.audio[0]
            self.Loader.FrameGenerator = self.Loader.StreamConainer.decode(audio=0)

            while not self.Loader._end.is_set():
                if self.Loader.Filter != self.Loader._Filter:
                    self.Loader._Filter = self.Loader.Filter

                    if self.Loader.Filter:
                        self.Loader.FilterGraph = AudioFilter()
                        self.Loader.FilterGraph.selectAudioStream = (
                            self.Loader.selectAudioStream
                        )
                        self.Loader.FilterGraph.setFilters(self.Loader.Filter)
                    else:
                        self.Loader.FilterGraph = None

                if not self.Resampler or self.Loader._haveToReloadResampler.is_set():
                    self.Resampler = av.AudioResampler(
                        format=av.AudioFormat("s16").packed,
                        layout="stereo" if CHANNELS >= 2 else "mono",
                        rate=SAMPLING_RATE,
                    )
                    self.Loader._haveToReloadResampler.clear()

                if self.Loader._seeking.locked():
                    self.Loader._seeking.acquire()
                    __seek_locked = True
                else:
                    __seek_locked = False

                Frame = next(self.Loader.FrameGenerator, None)

                if __seek_locked:
                    self.Loader._seeking.release()
                    self.Loader.AudioFifo.reset()

                if not Frame:
                    self.Loader.stop()
                    break

                _current = float(Frame.pts * self.Loader.selectAudioStream.time_base)

                if self.Loader.FilterGraph:
                    self.Loader.FilterGraph.push(Frame)
                    Frame = self.Loader.FilterGraph.pull()

                    if not Frame:
                        continue

                Frame.pts = None
                try:
                    Frame = self.Resampler.resample(Frame)
                except ValueError:
                    self.Loader.reloadResampler()
                    continue

                if not self.Loader.AudioFifo.haveToFillBuffer.is_set():
                    self.Loader.AudioFifo.haveToFillBuffer.wait()

                self.Loader.AudioFifo.write(Frame)
                self.Loader.current = _current

    def run(self):
        try:
            self._do_run()
        except:
            pass
        finally:
            if self.Loader.StreamConainer:
                self.Loader.StreamConainer.close()
                self.Loader.StreamConainer = None

            self.Loader.stop()



from logging import getLogger

from .utils import EventEmitter
from .voice_client import VoiceClient

log = getLogger("discodo.event")


class DiscordEvent:
    def __init__(self, client):
        self.client = client

        self.EventEmitter = EventEmitter()

        self.EventEmitter.on("READY", self.parseReady)
        self.EventEmitter.on("RESUME", self.parseResume)
        self.EventEmitter.on("VOICE_STATE_UPDATE", self.parseVoiceStateUpdate)
        self.EventEmitter.on("VOICE_SERVER_UPDATE", self.parseVoiceServerUpdate)

    async def emit(self, event, *args, **kwargs):
        return await self.EventEmitter.emit(event, *args, **kwargs)

    def dispatch(self, event, *args, **kwargs):
        return self.EventEmitter.dispatch(event, *args, **kwargs)

    async def parseReady(self, data):
        log.info(
            f'ready event dispatched. set session id {data["session_id"]} and user id {data["user"]["id"]}'
        )
        self.client.session_id = data["session_id"]
        self.client.user_id = int(data["user"]["id"])

    async def parseResume(self, data):
        log.info(
            f'resume event dispatched. set session id {data["session_id"]} and user id {data["user"]["id"]}'
        )
        self.client.session_id = data["session_id"]
        self.client.user_id = int(data["user"]["id"])

    async def parseVoiceStateUpdate(self, data):
        if not self.client.user_id or data["user_id"] != str(self.client.user_id):
            return
        log.info(f'recieve self voice update. set session id {data["session_id"]}')
        self.client.session_id = data["session_id"]

    async def parseVoiceServerUpdate(self, data):
        if int(data["guild_id"]) in self.client.voiceClients:
            log.info(
                f'Voice Client of {data["guild_id"]} found. reconnect to new endpoint'
            )

            vc = self.client.voiceClients[int(data["guild_id"])]
            await vc.createSocket(data)
        else:
            log.info(f'Voice Client of {data["guild_id"]} not found. create new one.')
            vc = VoiceClient(self.client, data)
            self.client.voiceClients[int(data["guild_id"])] = vc

        return vc



from xml.etree import ElementTree

from aiohttp import ClientSession
from markdownify import markdownify


class srv1:
    def __init__(self, Tree):
        self.Tree = Tree

        self.TextElements = {
            float(TextElement.attrib["start"]): {
                "start": float(TextElement.attrib["start"]),
                "duration": float(TextElement.attrib["dur"]),
                "end": round(
                    float(TextElement.attrib["start"])
                    + float(TextElement.attrib["dur"]),
                    2,
                ),
                "text": TextElement.text,
                "markdown": markdownify(TextElement.text).strip(),
            }
            for TextElement in self.Tree.findall("text")
        }

        self.duration = sorted(
            [TextElement["end"] for TextElement in self.TextElements.values()]
        )[-1]

        self.time = 0.0
        self.current = None

    @classmethod
    async def load(cls, URL: str):
        async with ClientSession() as session:
            async with session.get(URL) as session:
                Data = await session.text()

        Tree = ElementTree.fromstring(Data)
        return cls(Tree)

    def seek(self, time: float) -> str:
        self.time = float(time)

        if (
            not self.current
            or not self.current["start"] <= self.time < self.current["end"]
        ):
            Elements = [
                TextElement
                for Start, TextElement in self.TextElements.items()
                if Start <= self.time < TextElement["end"]
            ]
            self.current = Elements[-1] if Elements else None

        return self.lyrics

    @property
    def json(self) -> dict:
        return self.TextElements

    @property
    def lyrics(self) -> str:
        if self.current:
            return self.current
        return None

    @property
    def is_done(self) -> bool:
        return self.time >= self.duration



import audioop
import os

from ..exceptions import AudioSourceNotPlaying
from ..natives import AudioFifo, Loader

SAMPLING_RATE = int(os.getenv("SAMPLING_RATE", "48000"))
FRAME_LENGTH = int(os.getenv("FRAME_LENGTH", "20"))
SAMPLE_SIZE = int(os.getenv("SAMPLE_SIZE", "4"))
SAMPLES_PER_FRAME = int(SAMPLING_RATE / 1000 * FRAME_LENGTH)


class AudioSource:
    def __init__(self, file, volume=1.0, AudioData=None):
        self._volume = volume

        self.AudioFifo = AudioFifo()
        self.Loader = Loader(file, self.AudioFifo)
        self.Loader.start()

        self.AudioData = AudioData

        self.AVDurationLoaded = False

        self._filter = {}
        self._filterGraph = None

        self._duration = 0.0
        self.stopped = False

    def __del__(self):
        self.cleanup()

    def __getattr__(self, name: str):
        return getattr(self.AudioData, name)

    @property
    def volume(self) -> float:
        return self._volume

    @volume.setter
    def volume(self, value: float):
        self._volume = max(value, 0.0)

    @property
    def duration(self) -> float:
        return round(
            self.Loader.current
            - self.AudioFifo.samples
            / SAMPLES_PER_FRAME
            / 50
            / float(self._filter.get("atempo", "1.0")),
            2,
        )

    @property
    def remain(self) -> float:
        return round(self.AudioData.duration - self.duration, 2)

    @property
    def filter(self) -> dict:
        return self._filter

    @filter.setter
    def filter(self, value: dict):
        self._filter = value
        self.Loader.Filter = value

    def read(self) -> bytes:
        if not self.AudioFifo:
            return

        Data = self.AudioFifo.read()

        if not self.AVDurationLoaded and self.Loader and self.Loader.duration:
            self.AudioData.duration = self.Loader.duration
            self.AVDurationLoaded = True

        if not Data and self.Loader and self.Loader._buffering.locked():
            while self.Loader._buffering.locked():
                Data = self.AudioFifo.read()
                if Data:
                    break

        if Data and self.volume != 1.0:
            Data = audioop.mul(Data, 2, min(self._volume, 2.0))

        return Data

    def seek(self, offset: float):
        if not self.Loader:
            raise AudioSourceNotPlaying

        offset = (
            min(max(offset, 1), self.AudioData.duration - 1)
            if self.AudioData.duration
            else max(offset, 1)
        )
        self.Loader.seek(offset, any_frame=True)

    def stop(self):
        self.stopped = True
        return self.stopped

    def cleanup(self):
        self.Loader.stop()
        if self.AudioFifo and not self.AudioFifo.haveToFillBuffer.is_set():
            self.AudioFifo.haveToFillBuffer.set()
        self.AudioFifo = self.Loader.AudioFifo = None



import asyncio
import logging
from re import compile as Regex
from typing import Optional

from youtube_dl import YoutubeDL as YoutubeDLClient

from discodo.exceptions import NoSearchResults

log = logging.getLogger("discodo.extractor")

YOUTUBE_PLAYLIST_ID_REGEX = Regex(
    r"(?:http|https|)(?::\/\/|)(?:www.|)(?:music.|)(?:youtu\.be\/|youtube\.com(?:\/embed\/|\/v\/|\/watch\?v=|\/ytscreeningroom\?v=|\/feeds\/api\/videos\/|\/user\S*[^\w\-\s]|\S*[^\w\-\s]))([\w\-]{12,})[a-z0-9;:@#?&%=+\/\$_.-]*(?:&index=|)([0-9]*)?"
)


def _extract(query: str) -> Optional[dict]:
    option = {
        "format": "(bestaudio[ext=opus]/bestaudio/best)[protocol!=http_dash_segments]",
        "nocheckcertificate": True,
        "ignoreerrors": True,
        "no_warnings": True,
        "default_search": "auto",
        "source_address": "0.0.0.0",
        "logger": log,
        "skip_download": True,
        "writesubtitles": True,
    }

    YoutubePlaylistMatch = YOUTUBE_PLAYLIST_ID_REGEX.match(query)
    if YoutubePlaylistMatch and not YoutubePlaylistMatch.group(1).startswith(
        ("RD", "UL", "PU")
    ):
        option["playliststart"] = (
            int(YoutubePlaylistMatch.group(2))
            if YoutubePlaylistMatch.group(2).isdigit()
            else 1
        )
        option["dump_single_json"] = True
        option["extract_flat"] = True
        query = "https://www.youtube.com/playlist?list=" + YoutubePlaylistMatch.group(1)
    else:
        option["noplaylist"] = True

    YoutubeDL = YoutubeDLClient(option)
    Data = YoutubeDL.extract_info(query, download=False)

    if not Data:
        raise NoSearchResults

    if "entries" in Data:
        if len(Data["entries"]) == 1:
            return Data["entries"][0]

        return Data["entries"]

    return Data


def _clear_cache():
    option = {"ignoreerrors": True, "no_warnings": True, "logger": log}

    YoutubeDL = YoutubeDLClient(option)
    YoutubeDL.cache.remove()


async def extract(query, loop=None):
    if not loop:
        loop = asyncio.get_event_loop()

    return await loop.run_in_executor(None, _extract, query)


async def clear_cache(loop=None):
    if not loop:
        loop = asyncio.get_event_loop()

    return await loop.run_in_executor(None, _clear_cache)



from ...utils import EventEmitter


class VoiceClient:
    def __init__(self, Node, guild_id):
        self.Node = Node
        self.guild_id = guild_id

        self.emitter = EventEmitter()

    def __del__(self):
        vc = self.Node.voiceClients.get(self.guild_id)
        if vc and vc is self:
            self.Node.voiceClients.pop(self.guild_id)

    async def send(self, Operation: str, Data: dict = {}):
        Data["guild_id"] = self.guild_id

        return await self.Node.send(Operation, Data)

    async def loadSong(self, Query: str) -> dict:
        await self.send("loadSong", {"query": Query})

        return (await self.emitter.wait_for("loadSong", timeout=10.0))["song"]

    async def putSong(self, Song: dict) -> dict:
        await self.send("putSong", {"song": Song})

        return (await self.emitter.wait_for("putSong", timeout=10.0))["song"]

    async def skip(self, offset: int = 1) -> dict:
        await self.send("skip", {"offset": offset})

        return (await self.emitter.wait_for("skip", timeout=10.0))["remain"]

    async def seek(self, offset: float) -> dict:
        await self.send("seek", {"offset": offset})

        return await self.emitter.wait_for("seek", timeout=10.0)

    async def setVolume(self, volume: int) -> dict:
        await self.send("setVolume", {"volume": volume})

        return (await self.emitter.wait_for("setVolume", timeout=10.0))["volume"]

    async def setCrossfade(self, crossfade: float) -> dict:
        await self.send("setCrossfade", {"crossfade": crossfade})

        return (await self.emitter.wait_for("setCrossfade", timeout=10.0))["crossfade"]

    async def setAutoplay(self, autoplay: bool) -> dict:
        await self.send("setAutoplay", {"autoplay": autoplay})

        return (await self.emitter.wait_for("setAutoplay", timeout=10.0))["autoplay"]

    async def setFilter(self, filter: dict) -> dict:
        await self.send("setFilter", {"filter": filter})

        return await self.emitter.wait_for("setFilter", timeout=10.0)

    async def pause(self) -> dict:
        await self.send("pause")

        return await self.emitter.wait_for("pause", timeout=10.0)

    async def resume(self) -> dict:
        await self.send("resume")

        return await self.emitter.wait_for("resume", timeout=10.0)

    async def repeat(self, repeat: bool) -> dict:
        await self.send("repeat", {"repeat": repeat})

        return (await self.emitter.wait_for("repeat", timeout=10.0))["repeat"]

    async def changePause(self) -> dict:
        await self.send("changePause")

        return await self.emitter.wait_for("changePause", timeout=10.0)

    async def getQueue(self) -> dict:
        await self.send("getQueue")

        return (await self.emitter.wait_for("Queue", timeout=10.0))["entries"]

    async def getState(self) -> dict:
        await self.send("getState")

        return await self.emitter.wait_for("State", timeout=10.0)

    async def shuffle(self) -> dict:
        await self.send("shuffle")

        return await self.emitter.wait_for("shuffle", timeout=10.0)

    async def remove(self, index: int) -> dict:
        await self.send("remove", {"index": index})

        return await self.emitter.wait_for("remove", timeout=10.0)

    async def requestLyrics(self, language: int) -> dict:
        await self.send("requestLyrics", {"language": language})

        return await self.emitter.wait_for("requestLyrics", timeout=10.0)

    async def destroy(self) -> dict:
        await self.send("VC_DESTROY")

        return await self.emitter.wait_for("VC_DESTROYED", timeout=10.0)



import nacl.secret


def xsalsa20_poly1305(key, header, data):
    Box = nacl.secret.SecretBox(bytes(key))
    Nonce = bytearray(24)
    Nonce[:12] = header

    return header + Box.encrypt(bytes(data), bytes(Nonce)).ciphertext


def xsalsa20_poly1305_suffix(key, header, data):
    Box = nacl.secret.SecretBox(bytes(key))
    Nonce = nacl.utils.random(nacl.secret.SecretBox.NONCE_SIZE)

    return header + Box.encrypt(bytes(data), Nonce).ciphertext + Nonce


def getEncryptModes():
    return globals()



class DiscodoException(Exception):
    """Base exception class for Discodo
    """

    pass


class VoiceClientNotFound(DiscodoException):
    pass


class NoSearchResults(DiscodoException):
    pass


class AudioSourceNotPlaying(DiscodoException):
    pass


class NodeNotConnected(DiscodoException):
    pass



import asyncio
import os
import socket
import struct
import threading
from logging import getLogger

from websockets.exceptions import ConnectionClosed

from .encrypt import getEncryptModes
from .gateway import VoiceSocket
from .natives import opus

VCTIMEOUT = float(os.getenv("VCTIMEOUT", "300.0"))
SAMPLING_RATE = int(os.getenv("SAMPLING_RATE", "48000"))
FRAME_LENGTH = int(os.getenv("FRAME_LENGTH", "20"))
SAMPLES_PER_FRAME = int(SAMPLING_RATE / 1000 * FRAME_LENGTH)

log = getLogger("discodo.VoiceConnector")


class VoiceConnector:
    def __init__(self, client, data):
        self.loop = asyncio.get_event_loop()
        self.ws = None
        self.socket = None

        self.client = client
        self.user_id = self.client.user_id

        self.data = data
        self.session_id = self.client.session_id
        self.guild_id = self.data.get("guild_id")

        self._sequence = 0
        self._timestamp = 0

        self._connected = asyncio.Event()
        self._connectedThread = threading.Event()

        self._polling = None
        self.encoder = opus.Encoder()
        self.loop.create_task(self.createSocket())

    def __del__(self):
        if self.socket:
            try:
                self.socket.close()
            except:
                pass

        if self.ws:
            self.loop.create_task(self.ws.close(4000))

        if self._polling:
            self._polling.cancel()

    @property
    def sequence(self) -> int:
        return self._sequence

    @sequence.setter
    def sequence(self, value: int):
        if self._sequence + value > 65535:
            self._sequence = 0
        else:
            self._sequence = value

    @property
    def timestamp(self) -> int:
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: int):
        if self._timestamp + value > 4294967295:
            self._timestamp = 0
        else:
            self._timestamp = value

    async def createSocket(self, data: dict = None):
        self._connected.clear()
        self._connectedThread.clear()

        if data:
            self.data = data

        self.guild_id = self.data.get("guild_id")

        self.token = self.data.get("token")
        endpoint = self.data.get("endpoint")
        self.endpoint = endpoint.replace(":80", "")
        self.endpointIp = socket.gethostbyname(self.endpoint)
        log.info(f"voice endpoint {self.endpoint} ({self.endpointIp}) detected.")

        if self.socket:
            try:
                self.socket.close()
            except:
                pass

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setblocking(False)

        if self.ws:
            await self.ws.close(4000)

        if hasattr(self, "secretKey"):
            del self.secretKey

        self.ws = await VoiceSocket.connect(self)
        while not hasattr(self, "secretKey"):
            await self.ws.poll()

        if not self._polling or self._polling.done():
            self._polling = self.loop.create_task(self.pollingWs())

        self._connected.set()
        self._connectedThread.set()

    async def pollingWs(self):
        while True:
            try:
                await self.ws.poll()
            except (asyncio.TimeoutError, ConnectionClosed) as e:
                self._connected.clear()
                self._connectedThread.clear()

                if isinstance(e, ConnectionClosed):
                    reason = f"with {e.code}"
                else:
                    reason = "because timed out."

                log.info(
                    f"voice connection of {self.guild_id} destroyed {reason}. wait for events."
                )

                try:
                    await asyncio.wait_for(self._connected.wait(), timeout=VCTIMEOUT)
                except asyncio.TimeoutError:
                    return self.__del__()

    def makePacket(self, data: bytes) -> bytes:
        header = bytearray(12)
        header[0] = 0x80
        header[1] = 0x78

        struct.pack_into(">H", header, 2, self.sequence)
        struct.pack_into(">I", header, 4, self.timestamp)
        struct.pack_into(">I", header, 8, self.ssrc)

        encryptPacket = getEncryptModes()[self.encryptMode]
        return encryptPacket(self.secretKey, header, data)

    def send(self, data: bytes, encode: bool = True):
        self.sequence += 1
        if encode:
            data = self.encoder.encode(data)

        Packet = self.makePacket(data)

        self.socket.sendto(Packet, (self.endpointIp, self.endpointPort))
        self.timestamp += SAMPLES_PER_FRAME



import av
import av.filter


class AudioFilter:
    def __init__(self):
        self.selectAudioStream = None
        self._Filters = {}
        self._FilterChains = []
        self.Graph = None
        self.configured = False

    def setFilters(self, filters: dict):
        self._Filters = {}
        for filter, value in filters.items():
            self._Filters[filter] = value
        self._Filters["abuffersink"] = None

        self.configure()

    def configure(self):
        if not self.selectAudioStream:
            return

        self.Graph = av.filter.Graph()
        self._FilterChains = []
        self._FilterChains.append(self.Graph.add_abuffer(self.selectAudioStream))
        for filter, value in self._Filters.items():
            self._FilterChains.append(self.Graph.add(filter, value))
            self._FilterChains[-2].link_to(self._FilterChains[-1])
        self.Graph.configure()

        self.configured = True

    def push(self, Frame: av.AudioFrame):
        if not self.Graph:
            return

        return self.Graph.push(Frame)

    def pull(self) -> av.AudioFrame:
        if not self.Graph:
            return

        try:
            return self.Graph.pull()
        except av.error.BlockingIOError:
            return None



import threading


class withLock:
    def __init__(self, Lock: threading.Lock):
        self.Lock = Lock

    def __enter__(self):
        self.Lock.acquire()
        return self

    def __exit__(self, type, value, traceback):
        self.Lock.release()



# -*- coding: utf-8 -*-

__title__ = "discodo"
__author__ = "kijk2869"
__lisence__ = "MIT"
__version__ = "0.0.6a"

from collections import namedtuple

from .AudioSource import *
from .gateway import VoiceSocket
from .manager import AudioManager
from .natives import *
from .node import *
from .player import Player
from .stat import getStat
from .utils import *
from .voice_client import VoiceClient
from .voice_connector import VoiceConnector

try:
    import discord
except ModuleNotFoundError:
    pass
else:
    from .DPYClient import DPYClient

VersionInfo = namedtuple("VersionInfo", "major minor micro releaselevel serial")

version_info = VersionInfo(major=0, minor=0, micro=1, releaselevel="alpha", serial=0)



import asyncio
import collections
import functools
import traceback


class EventEmitter:
    def __init__(self, loop: asyncio.AbstractEventLoop = None):
        self._Events = collections.defaultdict(list)
        self._listeners = collections.defaultdict(list)
        self._Any = []

        self.loop = loop or asyncio.get_event_loop()

    def on(self, event: str, func):
        self._Events[event].append(func)
        return self

    def off(self, event: str, func):
        self._Events[event].remove(func)
        return self

    def onAny(self, func):
        self._Any.append(func)
        return self

    def offAny(self, func):
        self._Any.remove(func)
        return self

    def dispatch(self, event: str, *args, **kwargs):
        if self._Any:
            for func in self._Any:
                try:
                    if asyncio.iscoroutinefunction(func):
                        self.loop.create_task(func(event, *args, **kwargs))
                    else:
                        self.loop.call_soon_threadsafe(
                            functools.partial(func, event, *args, **kwargs)
                        )
                except:
                    traceback.print_exc()

        listeners = self._listeners.get(event)
        if listeners:
            for Future in listeners:
                if not (Future.done() or Future.cancelled()):
                    Future.set_result(*args, **kwargs)

                listeners.remove(Future)

            if not listeners:
                self._listeners.pop(event)

        if not event in self._Events:
            return

        for func in self._Events[event]:
            try:
                if asyncio.iscoroutinefunction(func):
                    self.loop.create_task(func(*args, **kwargs))
                else:
                    self.loop.call_soon_threadsafe(
                        functools.partial(func, *args, **kwargs)
                    )
            except:
                traceback.print_exc()

    def event(self, event: str):
        def wrapper(func):
            self.on(event, func)

        return wrapper

    def once(self, event: str, func):
        def wrapper(*args, **kwargs):
            self.off(event, func)
            return event(*args, **kwargs)

        return self.on(event, wrapper)

    def wait_for(self, event: str, timeout=None):
        Future = self.loop.create_future()

        self._listeners[event].append(Future)

        return asyncio.wait_for(Future, timeout)



import asyncio
import concurrent.futures
import json
import threading
import time
from collections import deque
from logging import getLogger

import websockets

log = getLogger("discodo.client")


class keepAlive(threading.Thread):
    def __init__(self, ws, interval: int, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        self.daemon = True

        self.ws = ws
        self.interval = interval
        self.Stopped = threading.Event()
        self.latency = None
        self.recent_latencies = deque(maxlen=20)

        self._lastAck = self._lastSend = time.perf_counter()
        self.timeout = ws.heartbeatTimeout
        self.threadId = ws.threadId

    def run(self):
        while not self.Stopped.wait(self.interval):
            if (self._lastAck + self.timeout) < time.perf_counter():
                Runner = asyncio.run_coroutine_threadsafe(
                    self.ws.close(4000), self.ws.loop
                )

                try:
                    Runner.result()
                except:
                    pass
                finally:
                    return self.stop()

            payload = {"op": "HEARTBEAT", "d": int(time.time() * 1000)}
            Runner = asyncio.run_coroutine_threadsafe(
                self.ws.sendJson(payload), self.ws.loop
            )
            try:
                totalBlocked = 0
                while True:
                    try:
                        Runner.result(10)
                        break
                    except concurrent.futures.TimeoutError:
                        totalBlocked += 10
                        log.warning(
                            f"Heartbeat blocked for more than {totalBlocked} seconds."
                        )
            except:
                return self.stop()
            else:
                self._lastSend = time.perf_counter()

    def ack(self):
        self._lastAck = time.perf_counter()
        self.latency = self._lastAck - self._lastSend
        self.recent_latencies.append(self.latency)

    def stop(self):
        self.Stopped.set()


class NodeConnection(websockets.client.WebSocketClientProtocol):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    async def connect(cls, client, loop=asyncio.get_event_loop()):
        ws = await websockets.connect(
            client.URL,
            loop=loop,
            klass=cls,
            extra_headers={"Authorization": client.password},
        )

        ws.client = client
        ws.heartbeatTimeout = 60.0
        ws.threadId = threading.get_ident()

        return ws

    @property
    def latency(self) -> float:
        return self._keepAliver.latency if self._keepAliver else None

    @property
    def averageLatency(self) -> float:
        if not self._keepAliver:
            return None

        return sum(self._keepAliver.recent_latencies) / len(
            self._keepAliver.recent_latencies
        )

    async def sendJson(self, data):
        log.debug(f"send to websocket {data}")
        await super().send(json.dumps(data))

    async def send(self, Operation, Data):
        payload = {"op": Operation, "d": Data}

        await self.sendJson(payload)

    async def poll(self):
        try:
            Message = await asyncio.wait_for(self.recv(), timeout=30.0)
            JsonData = json.loads(Message)
        except websockets.exceptions.ConnectionClosed as exc:
            raise exc
        else:
            Operation, Data = JsonData["op"], JsonData["d"]

            if Operation == "HELLO":
                await self.HELLO(Data)
            elif Operation == "HEARTBEAT_ACK":
                await self.HEARTBEAT_ACK(Data)

            return Operation, Data

    async def close(self, *args, **kwargs):
        if self._keepAliver:
            self._keepAliver.stop()

        await super().close(*args, **kwargs)

    async def HELLO(self, Data):
        self._keepAliver = keepAlive(self, min(Data["heartbeat_interval"], 5.0))
        self._keepAliver.start()

    async def HEARTBEAT_ACK(self, Data):
        self._keepAliver.ack()



import asyncio
import concurrent.futures
import json
import struct
import threading
import time
from collections import deque
from logging import getLogger

import websockets

from .encrypt import getEncryptModes

log = getLogger("discodo.gateway")


class keepAlive(threading.Thread):
    def __init__(self, ws, interval: int, *args, **kwargs):
        threading.Thread.__init__(self, *args, **kwargs)
        self.daemon = True

        self.ws = ws
        self.interval = interval
        self.Stopped = threading.Event()
        self.latency = None
        self.recent_latencies = deque(maxlen=20)

        self._lastAck = self._lastSend = time.perf_counter()
        self.timeout = ws.heartbeatTimeout
        self.threadId = ws.threadId

    def run(self):
        while not self.Stopped.wait(self.interval):
            if (self._lastAck + self.timeout) < time.perf_counter():
                Runner = asyncio.run_coroutine_threadsafe(
                    self.ws.close(4000), self.ws.loop
                )

                try:
                    Runner.result()
                except:
                    pass
                finally:
                    return self.stop()

            payload = {"op": self.ws.HEARTBEAT, "d": int(time.time() * 1000)}
            Runner = asyncio.run_coroutine_threadsafe(
                self.ws.sendJson(payload), self.ws.loop
            )
            try:
                totalBlocked = 0
                while True:
                    try:
                        Runner.result(10)
                        break
                    except concurrent.futures.TimeoutError:
                        totalBlocked += 10
                        log.warning(
                            f"Heartbeat blocked for more than {totalBlocked} seconds."
                        )
            except:
                return self.stop()
            else:
                self._lastSend = time.perf_counter()

    def ack(self):
        self._lastAck = time.perf_counter()
        self.latency = self._lastAck - self._lastSend
        self.recent_latencies.append(self.latency)

    def stop(self):
        self.Stopped.set()


class VoiceSocket(websockets.client.WebSocketClientProtocol):
    IDENTIFY = 0
    SELECT_PROTOCOL = 1
    READY = 2
    HEARTBEAT = 3
    SESSION_DESCRIPTION = 4
    SPEAKING = 5
    HEARTBEAT_ACK = 6
    RESUME = 7
    HELLO = 8
    RESUMED = 9
    CLIENT_DISCONNECT = 13

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    async def connect(cls, client, loop=asyncio.get_event_loop(), resume=False):
        ws = await websockets.connect(
            f"wss://{client.endpoint}/?v=4", loop=loop, klass=cls, compression=None
        )
        ws.client = client

        ws.heartbeatTimeout = 60.0
        ws.threadId = threading.get_ident()

        if not resume:
            await ws.identify()
        else:
            await ws.resume()

        return ws

    @property
    def latency(self):
        return self._keepAliver.latency if self._keepAliver else None

    @property
    def averageLatency(self):
        if not self._keepAliver:
            return None

        return sum(self._keepAliver.recent_latencies) / len(
            self._keepAliver.recent_latencies
        )

    async def sendJson(self, data):
        log.debug(f"send to websocket {data}")
        await self.send(json.dumps(data))

    async def identify(self):
        payload = {
            "op": self.IDENTIFY,
            "d": {
                "server_id": str(self.client.guild_id),
                "user_id": str(self.client.user_id),
                "session_id": self.client.session_id,
                "token": self.client.token,
            },
        }
        await self.sendJson(payload)

    async def resume(self):
        payload = {
            "op": self.RESUME,
            "d": {
                "server_id": str(self.client.guild_id),
                "user_id": str(self.client.user_id),
                "session_id": self.client.session_id,
                "token": self.client.token,
            },
        }
        await self.sendJson(payload)

    async def select_protocol(self, ip, port, mode):
        payload = {
            "op": self.SELECT_PROTOCOL,
            "d": {
                "protocol": "udp",
                "data": {"address": ip, "port": port, "mode": mode},
            },
        }
        await self.sendJson(payload)

    async def speak(self, state: bool = True):
        payload = {"op": self.SPEAKING, "d": {"speaking": int(state), "delay": 0}}
        await self.sendJson(payload)

    async def receive(self, message):
        Operation, Data = message["op"], message.get("d")
        log.debug(f"websocket recieved {Operation}: {Data}")

        if Operation == self.READY:
            await self.createConnection(Data)
        elif Operation == self.HEARTBEAT_ACK:
            self._keepAliver.ack()
        elif Operation == self.SESSION_DESCRIPTION:
            await self.loadKey(Data)
        elif Operation == self.HELLO:
            interval = Data["heartbeat_interval"] / 1000.0
            self._keepAliver = keepAlive(self, min(interval, 5.0))
            self._keepAliver.start()

    async def createConnection(self, data):
        self.client.ssrc = data["ssrc"]
        self.client.endpointPort = data["port"]
        self.client.endpointIp = data["ip"]

        packet = bytearray(70)
        struct.pack_into(">I", packet, 0, data["ssrc"])

        self.client.socket.sendto(
            packet, (self.client.endpointIp, self.client.endpointPort)
        )
        _recieved = await self.loop.sock_recv(self.client.socket, 70)

        _start, _end = 4, _recieved.index(0, 4)
        self.client.ip = _recieved[_start:_end].decode("ascii")
        self.client.port = struct.unpack_from(">H", _recieved, len(_recieved) - 2)[0]

        encryptModes = [
            Mode for Mode in data["modes"] if Mode in getEncryptModes().keys()
        ]
        log.debug(f'recieved encrypt modes {data["modes"]}')

        encryptMode = encryptModes[0]
        log.info(f"select encrypt mode {encryptMode}")

        await self.select_protocol(self.client.ip, self.client.port, encryptMode)

    async def loadKey(self, data):
        log.info("recieved voice secret key.")

        await self.speak(True)
        await self.speak(False)

        self.client.encryptMode = data["mode"]
        self.client.secretKey = data.get("secret_key")

    async def poll(self):
        try:
            Message = await asyncio.wait_for(self.recv(), timeout=30.0)
            await self.receive(json.loads(Message))
        except websockets.exceptions.ConnectionClosed as exc:
            raise exc

    async def close(self, *args, **kwargs):
        if self._keepAliver:
            self._keepAliver.stop()

        await super().close(*args, **kwargs)



import os
import threading

import psutil

Process = psutil.Process(os.getpid())


def getProcessMemory():
    return round(Process.memory_info().rss / 1e6)


def getProcessCpu():
    return Process.cpu_percent()


def getProcessThreads():
    return threading.active_count()


def getTotalCpu():
    return psutil.cpu_percent()


def getMemory():
    return round(psutil.virtual_memory().total / 1e6)


def getCpuCount():
    return psutil.cpu_count()


def getNetworkInbound():
    return round(psutil.net_io_counters().bytes_recv / 1e6)


def getNetworkOutbound():
    return round(psutil.net_io_counters().bytes_sent / 1e6)


def getStat():
    return {
        "UsedMemory": getProcessMemory(),
        "TotalMemory": getMemory(),
        "ProcessLoad": getProcessCpu(),
        "TotalLoad": getTotalCpu(),
        "Cores": getCpuCount(),
        "Threads": getProcessThreads(),
        "NetworkInbound": getNetworkInbound(),
        "NetworkOutbound": getNetworkOutbound(),
    }



import os
import threading

import av

SAMPLING_RATE = int(os.getenv("SAMPLING_RATE", "48000"))
FRAME_LENGTH = int(os.getenv("FRAME_LENGTH", "20"))
SAMPLE_SIZE = int(os.getenv("SAMPLE_SIZE", "4"))
SAMPLES_PER_FRAME = int(SAMPLING_RATE / 1000 * FRAME_LENGTH)

AUDIOBUFFERLIMIT = int(os.getenv("AUDIOBUFFERLIMIT", "5"))
AUDIOBUFFERLIMITMS = AUDIOBUFFERLIMIT * 50 * SAMPLES_PER_FRAME


class AudioFifo(av.AudioFifo):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.haveToFillBuffer = threading.Event()
        self.haveToFillBuffer.set()

    def read(self, samples: int = SAMPLES_PER_FRAME) -> bytes:
        AudioFrame = super().read(samples)
        if not AudioFrame:
            return

        if self.samples < AUDIOBUFFERLIMITMS:
            self.haveToFillBuffer.set()
        else:
            self.haveToFillBuffer.clear()

        return AudioFrame.planes[0].to_bytes()

    def write(self, *args, **kwargs):
        super().write(*args, **kwargs)

        if self.samples < AUDIOBUFFERLIMITMS:
            self.haveToFillBuffer.set()
        else:
            self.haveToFillBuffer.clear()

    def reset(self):
        super().read(samples=max(self.samples - 960, 0), partial=True)



import array
import ctypes
import os
import sys

SAMPLING_RATE = int(os.getenv("SAMPLING_RATE", "48000"))
CHANNELS = int(os.getenv("CHANNELS", "2"))
FRAME_LENGTH = int(os.getenv("FRAME_LENGTH", "20"))
SAMPLE_SIZE = int(os.getenv("SAMPLE_SIZE", "4"))
SAMPLES_PER_FRAME = int(SAMPLING_RATE / 1000 * FRAME_LENGTH)
FRAME_SIZE = SAMPLES_PER_FRAME * SAMPLES_PER_FRAME
EXPECTED_PACKETLOSS = int(os.getenv("EXPECTED_PACKETLOSS", "0"))
BITRATE = int(os.getenv("BITRATE", "128"))

_library = None

c_int_pointer = ctypes.POINTER(ctypes.c_int)
c_int16_pointer = ctypes.POINTER(ctypes.c_int16)
c_float_pointer = ctypes.POINTER(ctypes.c_float)


class EncoderStructure(ctypes.Structure):
    pass


EncoderStructurePointer = ctypes.POINTER(EncoderStructure)


def _errorLt(result, func, args):
    if result < 0:
        raise Exception
    return result


def _errorNe(result, func, args):
    ret = args[-1]._obj
    if ret.value != 0:
        raise Exception
    return result


exported_functions = {
    "opus_strerror": ([ctypes.c_int], ctypes.c_char_p, None),
    "opus_encoder_get_size": ([ctypes.c_int], ctypes.c_int, None),
    "opus_encoder_create": (
        [ctypes.c_int, ctypes.c_int, ctypes.c_int, c_int_pointer],
        EncoderStructurePointer,
        _errorNe,
    ),
    "opus_encode": (
        [
            EncoderStructurePointer,
            c_int16_pointer,
            ctypes.c_int,
            ctypes.c_char_p,
            ctypes.c_int32,
        ],
        ctypes.c_int32,
        _errorLt,
    ),
    "opus_encoder_ctl": (None, ctypes.c_int32, _errorLt),
    "opus_encoder_destroy": ([EncoderStructurePointer], None, None),
}


def loadOpus(name):
    global _library
    _library = loadLibopus(name)

    return _library is not None


def loadDefaultOpus():
    global _library
    try:
        if sys.platform == "win32":
            architecture = "x64" if sys.maxsize > 32 ** 2 else "x86"
            directory = os.path.dirname(os.path.abspath(__file__))
            _library = loadLibopus(
                os.path.join(directory, "bin", f"libopus-0.{architecture}.dll")
            )
        else:
            _library = loadLibopus(ctypes.util.find_library("opus"))
    except:
        _library = None

    return _library is not None


def loadLibopus(name):
    library = ctypes.cdll.LoadLibrary(name)

    for FuncName, Options in exported_functions.items():
        Func = getattr(library, FuncName)

        try:
            if Options[1]:
                Func.argtypes = Options[0]
            Func.restype = Options[1]
        except KeyError:
            pass

        try:
            if Options[2]:
                Func.errcheck = Options[2]
        except KeyError:
            pass

    return library


def isLoaded():
    global _library
    return _library is not None


ENCODER_CTL = {
    "OK": 0,
    "APPLICATION_AUDIO": 2049,
    "APPLICATION_VOIP": 2048,
    "APPLICATION_LOWDELAY": 2051,
    "CTL_SET_BITRATE": 4002,
    "CTL_SET_BANDWIDTH": 4008,
    "CTL_SET_FEC": 4012,
    "CTL_SET_PLP": 4014,
    "CTL_SET_SIGNAL": 4024,
}

BAND_CTL = {
    "narrow": 1101,
    "medium": 1102,
    "wide": 1103,
    "superwide": 1104,
    "full": 1105,
}

SIGNAL_CTL = {
    "auto": -1000,
    "voice": 3001,
    "music": 3002,
}


class Encoder:
    def __init__(self, application=ENCODER_CTL["APPLICATION_AUDIO"]):
        self.application = application

        if not isLoaded() and not loadDefaultOpus():
            raise ValueError

        self.state = self.createState()
        self.setBitrate(BITRATE)
        self.setFec(True)
        self.setExpectedPacketLoss(EXPECTED_PACKETLOSS)
        self.setBandwidth("full")
        self.setSignalType("auto")

    def createState(self):
        ret = ctypes.c_int()
        return _library.opus_encoder_create(
            SAMPLING_RATE, CHANNELS, self.application, ctypes.byref(ret)
        )

    def __del__(self):
        if hasattr(self, "state"):
            _library.opus_encoder_destroy(self.state)
            self.state = None

    def setBitrate(self, kbps):
        kbps = min(512, max(16, int(kbps)))

        _library.opus_encoder_ctl(
            self.state, ENCODER_CTL["CTL_SET_BITRATE"], kbps * 1024
        )

    def setBandwidth(self, req):
        if not req in BAND_CTL:
            raise KeyError

        _library.opus_encoder_ctl(
            self.state, ENCODER_CTL["CTL_SET_BANDWIDTH"], BAND_CTL[req]
        )

    def setSignalType(self, req):
        if not req in SIGNAL_CTL:
            raise KeyError

        _library.opus_encoder_ctl(
            self.state, ENCODER_CTL["CTL_SET_SIGNAL"], SIGNAL_CTL[req]
        )

    def setFec(self, enabled=True):
        _library.opus_encoder_ctl(
            self.state, ENCODER_CTL["CTL_SET_FEC"], 1 if enabled else 0
        )

    def setExpectedPacketLoss(self, percentage):
        _library.opus_encoder_ctl(self.state, ENCODER_CTL["CTL_SET_PLP"], percentage)

    def encode(self, Pcm, FrameSize=SAMPLES_PER_FRAME):
        max_length = len(Pcm)
        Pcm = ctypes.cast(Pcm, c_int16_pointer)
        Data = (ctypes.c_char * max_length)()

        Encoded = _library.opus_encode(self.state, Pcm, FrameSize, Data, max_length)

        return array.array("b", Data[:Encoded]).tobytes()



import asyncio
from logging import getLogger
from traceback import print_exc

import websockets

from discodo.exceptions import NodeNotConnected

from ...utils import EventEmitter
from .gateway import NodeConnection
from .voice_client import VoiceClient

log = getLogger("discodo.client")


class Node:
    def __init__(self, URL=None, password="hellodiscodo", user_id=None, reconnect=True):
        self.ws = None
        self.emitter = EventEmitter()

        self.loop = asyncio.get_event_loop()
        self.user_id = user_id
        self.reconnect = reconnect
        self.connected = asyncio.Event()

        self.URL = URL
        self.password = password

        self.voiceClients = {}
        self.emitter.onAny(self.onAnyEvent)

        self._polling = None
        self.loop.create_task(self.connect())

    async def connect(self):
        if self.ws and not self.ws.closed:
            await self.ws.close()

        self.ws = await NodeConnection.connect(self)
        self.voiceClients = {}
        self.connected.set()

        if not self._polling or self._polling.done():
            self._polling = self.loop.create_task(self.pollingWs())

        if self.user_id:
            await self.send("IDENTIFY", {"user_id": self.user_id})

    async def destroy(self):
        if self._polling and not self._polling.done():
            self._polling.cancel()

        if self.ws and not self.ws.closed:
            await self.ws.close()

        self.connected.clear()
        self.ws = None
        self.voiceClients = {}

    async def pollingWs(self):
        while True:
            try:
                Operation, Data = await self.ws.poll()
            except (asyncio.TimeoutError, websockets.ConnectionClosedError):
                self.connected.clear()
                if self.ws.closed and self.reconnect:
                    await self.connect()
                    continue

                if self.ws:
                    await self.ws.close()

                self.ws = None
                return
            else:
                log.debug(f"event {Operation} dispatched from websocket with {Data}")

                try:
                    self.emitter.dispatch(Operation, Data)
                except:
                    print_exc()

    def send(self, *args, **kwargs):
        if not self.ws:
            raise NodeNotConnected

        return self.ws.send(*args, **kwargs)

    async def onAnyEvent(self, Operation, Data):
        if Operation == "VC_CREATED":
            guild_id = int(Data["guild_id"])
            self.voiceClients[guild_id] = VoiceClient(self, guild_id)

        if Data and isinstance(Data, dict) and "guild_id" in Data:
            vc = self.getVC(Data["guild_id"])
            if vc:
                vc.emitter.dispatch(Operation, Data)

        if Operation == "VC_DESTROYED":
            guild_id = int(Data["guild_id"])
            if guild_id in self.voiceClients:
                self.voiceClients[guild_id].__del__()

    def getVC(self, guildID: int) -> VoiceClient:
        return self.voiceClients.get(int(guildID))

    async def discordDispatch(self, payload):
        if not payload["t"] in [
            "READY",
            "RESUME",
            "VOICE_STATE_UPDATE",
            "VOICE_SERVER_UPDATE",
        ]:
            return

        return await self.send("DISCORD_EVENT", payload)

    async def getStat(self) -> dict:
        await self.send("GET_STAT", None)

        return await self.emitter.wait_for("STAT", timeout=10.0)



import audioop
import math
import sys

import numpy as np


def volumeDetect(Data):
    volume = audioop.rms(Data, 2)
    return volume


def volumeNorm(Data):
    if not Data:
        return

    try:
        detected = volumeDetect(Data)
        sys.stdout.write("" + str(detected) + "|" + "#" * round(detected / 150) + "\n")
        a = 7000 / detected
        print(a)
        _Data = audioop.mul(Data, 2, a)
        b = volumeDetect(Data)
        sys.stdout.write("->" + str(b) + "|" + "#" * round(b / 150) + "\n")
        return _Data
    except:
        return Data



import aiohttp

from ..extractor import clear_cache, extract
from .AudioSource import AudioSource


class AudioData:
    def __init__(self, data):
        self.id = data.get("id")
        self.title = data.get("title")

        if data.get("_type") == "url" and data.get("ie_key") == "Youtube":
            self.webpage_url = f"https://www.youtube.com/watch?v={self.id}"
            self.thumbnail = f"https://i.ytimg.com/vi/{self.id}/hqdefault.jpg"
            self.stream_url = None
        else:
            self.webpage_url = data.get("webpage_url")
            self.thumbnail = data.get("thumbnail")
            self.stream_url = data.get("url")

        self.lyrics = (
            {
                lang: [
                    LyricsData["url"]
                    for LyricsData in lyrics
                    if LyricsData["ext"] == "srv1"
                ][0]
                for lang, lyrics in data["subtitles"].items()
            }
            if "subtitles" in data
            else {}
        )
        self.duration = data.get("duration")
        self.is_live = data.get("is_live", False)

        self.uploader = data.get("uploader")
        self.description = data.get("description")
        self.subtitles = data.get("subtitles")

        self.playlist = data.get("playlist")

        self._source = None

    @classmethod
    async def create(cls, query: str):
        Data = await extract(query)

        if isinstance(Data, list):
            return [cls(Item) for Item in Data]

        return cls(Data)

    async def gather(self):
        Data = await extract(self.webpage_url)
        self.__init__(Data)

        return self

    async def source(self, *args, _retry: int = 0, **kwargs) -> AudioSource:
        if not self.stream_url:
            await self.gather()

        if not self._source:
            async with aiohttp.ClientSession() as session:
                async with session.get(self.stream_url) as request:
                    Status = request.status

            if Status == 403:
                if _retry == 0:
                    return await self.source(*args, _retry=_retry + 1, **kwargs)
                elif _retry == 1:
                    await clear_cache()
                    return await self.source(*args, _retry=_retry + 1, **kwargs)

            self._source = AudioSource(self.stream_url, *args, AudioData=self, **kwargs)

        return self._source

    def toDict(self) -> dict:
        return {
            "id": self.id,
            "title": self.title,
            "webpage_url": self.webpage_url,
            "thumbnail": self.thumbnail,
            "duration": self.duration,
            "is_live": self.is_live,
            "uploader": self.uploader,
            "description": self.description,
            "playlist": self.playlist,
        }

    @classmethod
    def fromDict(cls, Data):
        return cls(Data)



import asyncio
import json
import logging
import os

from sanic import Blueprint
from sanic.websocket import ConnectionClosed

from ...manager import AudioManager
from .events import WebsocketEvents

WSINTERVAL = float(os.getenv("WSINTERVAL", "15"))
WSTIMEOUT = float(os.getenv("WSTIMEOUT", "60"))
PASSWORD = os.getenv("PASSWORD", "hellodiscodo")

log = logging.getLogger("discodo.server")


app = Blueprint(__name__)


@app.websocket("/")
async def feed(request, ws):
    handler = WebsocketHandler(request, ws)
    await handler.join()


class WebsocketHandler:
    def __init__(self, request, ws):
        self.loop = asyncio.get_event_loop()
        self.request = request
        self.ws = ws

        self.AudioManager = None

        self.loop.create_task(self.handle())
        self._running = asyncio.Event()

    def __del__(self):
        if self.AudioManager:
            self.AudioManager.__del__()
            self.AudioManager = None

    async def join(self):
        return await self._running.wait()

    async def handle(self):
        if self._running.is_set():
            self._running.clear()

        try:
            await self._handle()
        finally:
            self._running.set()
            self.__del__()

    async def _handle(self):
        log.info(f"new websocket connection created from {self.request.ip}.")

        if self.request.headers.get("Authorization") != PASSWORD:
            log.warning(
                f"websocket connection from {self.request.ip} forbidden: password mismatch."
            )
            await self.forbidden("Password mismatch.")
            return

        await self.hello()

        while True:
            try:
                RAWDATA = await asyncio.wait_for(self.ws.recv(), timeout=WSTIMEOUT)
            except (asyncio.TimeoutError, ConnectionClosed) as exception:
                if isinstance(exception, asyncio.TimeoutError):
                    log.info("websocket connection closing because of timeout.")
                elif isinstance(exception, ConnectionClosed):
                    log.info(
                        f"websocket connection disconnected. code {exception.code}"
                    )

                return

            try:
                _Data = json.loads(RAWDATA)
                Operation, Data = _Data.get("op"), _Data.get("d")
            except:
                continue

            if Operation and hasattr(WebsocketEvents, Operation):
                log.debug(f"{Operation} dispatched with {Data}")

                Func = getattr(WebsocketEvents, Operation)
                self.loop.create_task(Func(self, Data))

    async def sendJson(self, Data):
        log.debug(f"send {Data} to websocket connection of {self.request.ip}.")
        await self.ws.send(json.dumps(Data))

    async def initialize_manager(self, user_id):
        self.AudioManager = AudioManager(user_id=user_id)
        self.AudioManager.event.onAny(self.manager_event)

    async def manager_event(self, guild_id, Event, **kwargs):
        payload = {"op": Event, "d": {"guild_id": guild_id}}
        payload["d"] = dict(payload["d"], **kwargs)

        await self.sendJson(payload)

    async def hello(self):
        payload = {"op": "HELLO", "d": {"heartbeat_interval": WSINTERVAL}}

        await self.sendJson(payload)

    async def forbidden(self, message):
        payload = {"op": "FORBIDDEN", "d": message}

        await self.sendJson(payload)



import asyncio
import audioop
import os
import threading
import time
import traceback

from discodo.AudioSource.AudioSource import SAMPLES_PER_FRAME

from .AudioSource import AudioData, AudioSource

PRELOAD_TIME = int(os.getenv("PRELOAD_TIME", "10"))
FRAME_LENGTH = int(os.getenv("FRAME_LENGTH", "20"))
DELAY = FRAME_LENGTH / 1000.0


class Player(threading.Thread):
    def __init__(self, voice_client, volume=1.0, loop=None):
        threading.Thread.__init__(self)
        self.daemon = True
        self._end = threading.Event()
        self.loop = loop or asyncio.get_event_loop()

        self.client = voice_client
        self.speakState = False
        self.__next_called = False
        self.__current_future = None

        self._volume = volume

    @property
    def current(self) -> AudioSource:
        Source = self.client.InternalQueue[0] if self.client.InternalQueue else None
        if Source:
            if isinstance(Source, AudioData):
                if not self.__current_future:
                    self.__current_future = asyncio.run_coroutine_threadsafe(
                        Source.source(), self.loop
                    )
                    Source = None
                elif self.__current_future.done():
                    if self.client.InternalQueue[0] == Source:
                        Source = self.client.InternalQueue[
                            0
                        ] = self.__current_future.result()

                    self.__current_future = None

                    if Source.volume != 1.0:
                        Source.volume = 1.0

                    if not hasattr(Source, "_dispatched"):
                        Source._dispatched = True
                        self.client.event.dispatch(
                            "SongStart", song=Source.AudioData.toDict()
                        )
                else:
                    Source = None
            elif (
                isinstance(Source, AudioSource)
                and Source.filter != self.client.filter
                and hasattr(Source.Loader, "selectAudioStream")
            ):
                Source.filter = self.client.filter

        return Source

    @property
    def next(self) -> AudioSource:
        Source = (
            self.client.InternalQueue[1]
            if self.client.InternalQueue and len(self.client.InternalQueue) > 1
            else None
        )

        if Source and isinstance(Source, AudioSource):
            if not hasattr(Source, "_dispatched"):
                Source._dispatched = True
                self.client.event.dispatch("SongStart", song=Source.AudioData.toDict())
            if Source.filter != self.client.filter and hasattr(
                Source.Loader, "selectAudioStream"
            ):
                Source.filter = self.client.filter

        return Source

    def nextReady(self):
        self.loop.create_task(self._nextReady())

    async def _nextReady(self):
        Source = (
            self.client.InternalQueue[1]
            if self.client.InternalQueue and len(self.client.InternalQueue) > 1
            else None
        )
        if Source:
            if isinstance(Source, AudioData):
                Data = await Source.source()

                if self.client.InternalQueue[1] == Source:
                    Source = self.client.InternalQueue[1] = Data

                if Data.volume != 0.0:
                    Data.volume = 0.0
            elif (
                isinstance(Source, AudioSource)
                and Source.filter != self.client.filter
                and hasattr(Source.Loader, "selectAudioStream")
            ):
                Source.filter = self.client.filter

    def makeFrame(self) -> bytes:
        if not self.current:
            return

        Data = self.current.read()

        if not Data or self.current.volume == 0.0:
            self.loop.call_soon_threadsafe(self.current.cleanup)
            self.client.event.dispatch("SongEnd", song=self.current.AudioData.toDict())
            del self.client.InternalQueue[0]

        if (
            self.__next_called
            and self.current
            and not (
                self.current.remain <= (PRELOAD_TIME + self.client.crossfade)
                or self.current.stopped
            )
        ):
            self.__next_called = False

        if self.next and (
            self.current.remain <= (PRELOAD_TIME + self.client.crossfade)
            or self.current.stopped
        ):
            if self.__next_called:
                self.__next_called = False
            if isinstance(self.next, AudioData) and not hasattr(self.next, "_called"):
                self.next._called = True
                self.nextReady()
            elif (
                self.client.crossfade
                and isinstance(self.next, AudioSource)
                and (self.next.AudioFifo.samples > SAMPLES_PER_FRAME)
                and (
                    self.current.remain <= self.client.crossfade or self.current.stopped
                )
            ):
                NextData = self.next.read()

                if NextData:
                    CrossFadeVolume = 1.0 / (self.client.crossfade / DELAY)
                    if self.next.volume < 1.0:
                        self.next.volume = round(self.next.volume + CrossFadeVolume, 10)
                    if self.current.volume > 0.0:
                        self.current.volume = round(
                            self.current.volume - CrossFadeVolume, 10
                        )

                    Data = audioop.add(Data, NextData, 2)
        elif not self.__next_called and (
            self.current.remain <= (PRELOAD_TIME + self.client.crossfade)
            or self.current.stopped
        ):
            self.client.event.dispatch(
                "NeedNextSong", current=self.current.AudioData.toDict()
            )
            self.__next_called = True
        elif self.current and self.current.stopped:
            if isinstance(self.current, AudioSource) and self.current.volume > 0.0:
                CrossFadeVolume = 1.0 / (self.client.crossfade / DELAY)
                self.current.volume = round(self.current.volume - CrossFadeVolume, 10)
        else:
            if isinstance(self.current, AudioSource) and self.current.volume < 1.0:
                self.current.volume = round(self.current.volume + 0.01, 3)
            if isinstance(self.next, AudioSource) and self.next.volume != 0.0:
                self.next.volume = 0.0

        return Data

    def _do_run(self):
        self.loops = 0
        _start = time.perf_counter()

        while not self._end.is_set():
            try:
                if not self.client._connectedThread.is_set():
                    self.client._connectedThread.wait()
                    self.loops = 0
                    _start = time.perf_counter()

                if not self.client.paused:
                    Data = self.makeFrame()
                else:
                    Data = None

                if self._volume != self.client.volume:
                    if self._volume < self.client.volume:
                        self._volume = round(self._volume + 0.01, 3)
                    if self._volume > self.client.volume:
                        self._volume = round(self._volume - 0.01, 3)

                if Data:
                    if not self.speakState:
                        self.speak(True)

                    if self._volume != 1.0:
                        Data = audioop.mul(Data, 2, min(self._volume, 2.0))

                    self.client.send(Data)
                elif self.speakState:
                    self.speak(False)

                self.loops += 1
                nextTime = _start + DELAY * self.loops
                time.sleep(max(0, DELAY + (nextTime - time.perf_counter())))
            except:
                traceback.print_exc()

    def run(self):
        try:
            self._do_run()
        except:
            pass
        finally:
            self.stop()

    def stop(self):
        self._end.set()
        self.speak(False)

    def speak(self, value):
        self.speakState = value
        asyncio.run_coroutine_threadsafe(self.client.ws.speak(value), self.client.loop)



import asyncio

from discodo.exceptions import VoiceClientNotFound

from .AudioSource import AudioData
from .event import DiscordEvent
from .utils import EventEmitter
from .voice_client import VoiceClient


class AudioManager:
    def __init__(self, *args, **kwargs):
        self.loop = asyncio.get_event_loop()

        self.user_id = kwargs.get("user_id")
        self.session_id = kwargs.get("session_id")

        self.event = EventEmitter()

        self.voiceClients = {}

        self.discordEvent = DiscordEvent(self)

    def __del__(self):
        for voiceClient in self.voiceClients.values():
            self.loop.call_soon_threadsafe(voiceClient.__del__)

    def discordDispatch(self, data: dict):
        Event, Data = data["t"], data["d"]

        return self.discordEvent.dispatch(Event, Data)

    def getVC(self, guildID: int) -> VoiceClient:
        if not int(guildID) in self.voiceClients:
            raise VoiceClientNotFound

        return self.voiceClients.get(int(guildID))

    def delVC(self, guildID: int):
        self.getVC(guildID).__del__()

    async def getSong(self, Query: str) -> AudioData:
        return await AudioData.create(Query)

    async def putSong(self, guildID: int, *args, **kwargs) -> AudioData:
        return await self.getVC(guildID).putSong(*args, **kwargs)

    async def loadSong(self, guildID: int, *args, **kwargs) -> AudioData:
        return await self.getVC(guildID).loadSong(*args, **kwargs)

    def skip(self, guildID: int, offset: int):
        return self.getVC(guildID).skip(offset)

    def seek(self, guildID: int, offset: int):
        return self.getVC(guildID).seek(offset)

    def setVolume(self, guildID: int, value: float) -> float:
        self.getVC(guildID).volume = value

        return self.getVC(guildID).volume

    def setCrossfade(self, guildID: int, value: float) -> float:
        self.getVC(guildID).crossfade = value

        return self.getVC(guildID).crossfade

    def setAutoplay(self, guildID: int, value: bool) -> bool:
        self.getVC(guildID).autoplay = value

        return self.getVC(guildID).autoplay



import os
from functools import wraps

from sanic import Sanic, response
from sanic.exceptions import abort

from ...AudioSource import AudioData
from ...stat import getStat
from .websocket import app as WebsocketBlueprint

PASSWORD = os.getenv("PASSWORD", "hellodiscodo")

app = Sanic(__name__)

app.register_blueprint(WebsocketBlueprint)


def authorized(func):
    def wrapper(request, *args, **kwargs):
        if request.headers.get("Authorization") != PASSWORD:
            abort(403, "Password mismatch.")

        return func(request, *args, **kwargs)

    return wrapper


@app.route("/stat")
async def stat(request):
    return response.json(getStat())


@app.route("/getSong")
@authorized
async def getSong(request):
    query = "".join(request.args.get("query", [])).strip()

    if not query:
        abort(400, "Missing parameter query.")

    return response.json(await AudioData.create(query))
